/**
  Simple wrapper around operating system's
  memory manipulation routines.
  
  Orion Sky Lawlor, olawlor@acm.org, 2005/3/23 (Public Domain)
*/
#include "mmap.h"

#ifdef _WIN32
/** Windows Version */
#include <windows.h>
#include <string>

memory_mapping::memory_mapping(void *req_addr,int req_len) 
{
	/* If specifying an address, *must* reserve
	  address range before putting data there. */
	void *loc1=VirtualAlloc(req_addr,req_len,
		MEM_RESERVE,
		PAGE_NOACCESS);
	/* Now actually put memory at our address */
	addr=VirtualAlloc(req_addr,req_len,
		MEM_COMMIT,
		PAGE_READWRITE);
	if (addr==NULL) { printf("Error %d in VirtualAlloc!\n", GetLastError()); exit(1);}
	len=req_len;
}

memory_mapping::~memory_mapping() {
	VirtualFree(addr,len,MEM_RELEASE);
}

/** Show the current set of mappings. */
void memory_mapping::show(void) {
	char *addr=(char *)0x00000000;
	int pageSize=4096;
	do {
		MEMORY_BASIC_INFORMATION m;
		int len=VirtualQuery(addr,&m,sizeof(m));
		if (len==0) return; /* Kernel region */
		std::string accessType="free";
		
		int ap=m.AllocationProtect;
		if (ap&0x22) accessType="read";
		if (ap&0x44) accessType="read/write";
		if (ap&0x88) accessType="copy-on-write";
		if (ap&0xf0) accessType+=" execute";
		
		if (ap&PAGE_GUARD) accessType+=" guard";
		if (ap&PAGE_NOCACHE) accessType+=" nocache";
		if (ap&PAGE_WRITECOMBINE) accessType+=" writecombine";
		
		if (m.Type&MEM_IMAGE) accessType+=" image";
		if (m.Type&MEM_MAPPED) accessType+=" mapped";
		if (m.Type&MEM_PRIVATE) accessType+=" private";
		
		printf("%p - %p %5d KB: %s\n",
			m.BaseAddress,
			((char *)m.BaseAddress)+m.RegionSize,
			m.RegionSize>>10,
			accessType.c_str());
		addr+=m.RegionSize;
	} while (addr!=0);
}

#else
/** UNIX Version */
#include <stdio.h>     /* for printf */
#include <stdlib.h>     /* for exit */
#include <sys/mman.h>  /* for mmap */
#include <unistd.h>
#include <errno.h>     /* for perror */
#include <sys/types.h> /* for open */
#include <sys/stat.h>
#include <fcntl.h>

/**
  Memory must be mapped from a file-- use /dev/zero
  to get as much zero'd memory as you want.
*/
static int getZeroFd(void) {
	static int fd=0;
	if (fd==0) {
		fd=open("/dev/zero",O_RDWR,00700);
		if (fd<=0) {
			perror("Error during /dev/zero open");
			exit(1);
		}
	}
	return fd;
}
	
/** 
  Make a memory mapping at this address covering
  this many bytes. 
*/
memory_mapping::memory_mapping(void *req_addr,int req_len) {
	addr=mmap(req_addr, /* Address to map to (must be pagesize multiple!) */
		req_len, /* Data to map (will be rounded up to pagesize) */
		PROT_READ|PROT_WRITE|PROT_EXEC, /* Access control */
		MAP_FIXED|MAP_PRIVATE, /* Map flags */
		getZeroFd(), /* File to map */
		0); /* File offset (bytes) */
	if (addr==(void *)MAP_FAILED) {
		perror("error during mmap:");
		printf("Error during mmap at %p of %d bytes\n",
			req_addr,req_len);
		exit(1);
	}
	len=req_len;
}

/** Release this memory mapping. */
memory_mapping::~memory_mapping() {
	munmap(addr,len);
}

/** Show the current set of mappings. */
void memory_mapping::show(void) {
	char cmd[100];
	sprintf(cmd,"cat /proc/%d/maps",(int)getpid());
	system(cmd);
}
#endif


